from typing import List, BinaryIO, Optional, Dict
from struct import Struct

import numpy as np

from .blob import BlobSlice
from .label import Label
from .field import Field, normalize, serialize
from .derivation import Derivation
from .channel_base import ChannelBase


class Channel(ChannelBase):
    fields = [
        Field('label', str, 16),
        Field('channel_type', str, 80),
        Field('physical_dimension', str, 8),
        Field('physical_minimum', float, 8),
        Field('physical_maximum', float, 8),
        Field('digital_minimum', int, 8),
        Field('digital_maximum', int, 8),
        Field('prefiltering', str, 80),
        Field('num_samples_per_record', int, 8),
        Field('reserved', str, 32)
    ]

    def __init__(self, *args, **kwargs):
        self.signal: Optional[BlobSlice] = None
        for k, v in kwargs.items():
            getattr(type(self), k).fset(self, v)

    def __getitem__(self, sli: slice) -> np.ndarray:
        """return a slice of the signal"""
        if self.signal is None:
            raise RuntimeError(f"channel {self} uninitialized")

        scale = (self.physmax - self.physmin) / (self.digimax - self.digimin)
        offset = self.physmax / scale - self.digimax
        return scale * (self.signal[sli] + offset)

    def from_dict(self, signals: Dict[Label, np.ndarray]) -> np.ndarray:
        return signals[self.label]

    @property
    def channel_type(self) -> str:
        return self._channel_type

    @channel_type.setter
    def channel_type(self, v: str):
        self._channel_type = normalize(str, v)

    @property
    def physical_dimension(self) -> str:
        return self._physical_dimension

    @physical_dimension.setter
    def physical_dimension(self, v: str):
        self._physical_dimension = normalize(str, v)

    @property
    def physical_minimum(self) -> float:
        return self.physmin

    @physical_minimum.setter
    def physical_minimum(self, v: float):
        self.physmin = v

    @property
    def physical_maximum(self) -> float:
        return self.physmax

    @physical_maximum.setter
    def physical_maximum(self, v: float):
        self.physmax = v

    @property
    def digital_minimum(self) -> int:
        return self.digimin

    @digital_minimum.setter
    def digital_minimum(self, v: int):
        self.digimin = v

    @property
    def digital_maximum(self) -> int:
        return self.digimax

    @digital_maximum.setter
    def digital_maximum(self, v: int):
        self.digimax = v

    @property
    def prefiltering(self) -> str:
        return self._prefiltering

    @prefiltering.setter
    def prefiltering(self, v: str):
        self._prefiltering = normalize(str, v)

    @property
    def num_samples_per_record(self) -> int:
        return self._num_samples_per_record

    @num_samples_per_record.setter
    def num_samples_per_record(self, v: int):
        self._num_samples_per_record = v

    @property
    def reserved(self) -> str:
        return self._reserved

    @reserved.setter
    def reserved(self, v: str):
        self._reserved = normalize(str, v)

    def derive(self, other: ChannelBase) -> Derivation:
        if not self.is_compatible(other):
            raise ValueError(f"cannot derive {self} and {other}")

        return Derivation(self, other)

    @property
    def children(self) -> List[Label]:
        return [self.label]

    @classmethod
    def read(cls, file: BinaryIO, num_channels: int) -> List['Channel']:
        channels = [cls() for _ in range(num_channels)]
        for field in cls.fields:
            data = file.read(field.size * num_channels)
            format_str = (str(field.size) + 's') * num_channels
            values = Struct(format_str).unpack(data)
            normalized = (normalize(field.type, v) for v in values)
            for c, v in zip(channels, normalized):
                setattr(c, field.name, v)

        return channels

    @classmethod
    def write(cls, file: BinaryIO, channels: List['Channel']):
        serialized = b''
        for field in cls.fields:
            serialized += b''.join([
                serialize(getattr(channel, field.name), field.size)
                for channel in channels
            ])

        file.write(serialized)



from typing import Optional, Tuple
from .cached_property import cached_property
from .notation import synonyms


class Label(str):
    def __init__(self, original: str):
        """initialize Label with `.original` storing input str"""
        self.original = original

    def __new__(cls, original: str):
        """normalize original input str in new instance"""
        normalized = cls.normalize(original)
        return super().__new__(cls, normalized)

    @classmethod
    def normalize(cls, original: str) -> str:
        split = original.replace('/', '-').upper().split(' ')
        label = split[1] if 1 < len(split) else split[0]
        synonym_tuple = (
            synonyms.get(part, part)
            for part in label.split('-')
        )
        return '-'.join(synonym_tuple)

    @property
    def left(self) -> Optional[str]:
        return self.parts[0]

    @cached_property
    def right(self) -> Optional[str]:
        return self.parts[1]

    @cached_property
    def parts(self):
        parts = self.split('-')
        if len(parts) == 1:
            return parts + [None]
        elif parts[0] == '':
            parts[0] = None

        return parts

    def derive(self, other: 'Label') -> Tuple['Label', str]:
        if not self.is_compatible(other):
            raise ValueError(f"Unable to derive {self} with {other}")

        try:
            return self + other, '+'
        except ValueError:
            return self - other, '-'

    def __add__(self, other: 'Label') -> 'Label':  # type: ignore
        if self.right == other.left:
            return self.construct(self.left, other.right)
        elif self.left == other.right:
            return self.construct(other.left, self.right)

        raise ValueError(f"Can't add {self} and {other}")

    def __sub__(self, other: 'Label') -> 'Label':
        if self.left == other.left:
            return self.construct(other.right, self.right)
        elif self.right == other.right:
            return self.construct(self.left, other.left)

        raise ValueError(f"Can't subtract {other} from {self}")

    def is_compatible(self, o: 'Label') -> bool:
        p = self.parts
        return o.left in p or o.right in p

    def __neg__(self):
        return self.construct(self.right, self.left)

    @classmethod
    def construct(cls, ls: Optional[str], rs: Optional[str]) -> 'Label':
        left: str = ls or ''
        right: str = f"-{rs}" if rs else ''
        return cls(f"{left}{right}")



from typing import Dict

channel_labels_by_type = dict(
    REF=[
        'A1', 'A2', 'M1', 'M2'
    ],
    EEG=[
        'C3', 'C4', 'CZ',
        'EEG',
        'F1', 'F2', 'F3', 'F4', 'F7', 'F8', 'FZ', 'Fp1', 'Fp2', 'FpZ',
        'O1', 'O2', 'OZ',
        'P3', 'P4', 'PZ',
        'T3', 'T4', 'T5', 'T6'
    ],

    ECG=[
        'ECG', 'ECG1', 'ECG2'
    ],

    EMG=[
        'EMG',
        'EMG1', 'EMG2',
        'EMG_x', 'EMG_y',
        'CHIN',
        'CHIN1', 'CHIN2',
        'Deltoide',
        'Submental1', 'Submental2', 'Submental3',
    ],

    EOG=[
        'EOG',
        'EOG_L', 'EOG_R',
        'E1-M2', 'E2-M2',
        'EOG_L-M2', 'EOG_R-M2',
    ],

    HR=[
        'HR'
    ],

    RESP=[
        'Pleth', 'Thorax', 'Abdomen', 'Flow', 'Cannula', 'THE'
    ],

    POS=[
        'Position'
    ],

    BLOODGAS=[
        'SpO2', 'STAT_SpO2'
    ],

    SNORE=[
        'MIC'
    ],

    LEG=[
        'Leg',
        'TIB_L', 'TIB_R',
        'TIB_L1', 'TIB_R1',
    ]
)


synonyms = {
    'LEOG-REF': 'EOG_L',
    'REOG-REF': 'EOG_R',
    'E1': 'EOG_L',
    'E2': 'EOG_R',
    'E1-M2': 'EOG_L-M2',
    'E2-M2': 'EOG_R-M2',
    'E1M2': 'EOG_L-M2',
    'E2M1': 'EOG_R-M1',
    'E2M2': 'EOG_R-M2',
    'EOGLOC-A2': 'EOG_L-A2',
    'EOGROC-A1': 'EOG_R-A1',
    'A1': 'M1',
    'A2': 'M2',
    'EEGF3-A2': 'F3-M2',
    'EEGF4-A1': 'F4-M1',
    'EEGC3-A2': 'C3-M2',
    'EEGC4-A1': 'C4-M1',
    'EEGO1-A2': 'O1-M2',
    'EEGO2-A1': 'O2-M1',
    'EEGA1-A2': 'A1-M2',
    'EEGA2-A1': 'A2-M1',
    'EOGLEFT': 'EOG_L-A2',
    'EOGRIGHT': 'EOG_R-A2',

    'F3A2': 'F3-M2',
    'F4A1': 'F4-M1',
    'C3A2': 'C3-M2',
    'C4A1': 'C4-M1',
    'O1A2': 'O1-M2',
    'O2A1': 'O2-M1',

    'F3-REF': 'F3',
    'F4-REF': 'F4',
    'C3-REF': 'C3',
    'C4-REF': 'C4',
    'O1-REF': 'O1',
    'O2-REF': 'O2',
    'M1-REF': 'M1',
    'M2-REF': 'M2',

    'F3M2': 'F3-M2',
    'F4M1': 'F4-M1',
    'C3M2': 'C3-M2',
    'C4M1': 'C4-M1',
    'O1M2': 'O1-M2',
    'O2M1': 'O2-M1',
    'FP1': 'Fp1',
    'FP2': 'Fp2',
    'FPZ': 'FpZ',
    'EKG': 'ECG',
    'EKG1': 'ECG1',
    'EKG2': 'ECG2',
    'EMG_X': 'EMG_x',
    'EMG_Y': 'EMG_y',
    'EMG-EMG': 'EMG_x-EMG_y',
    'MILO': 'CHIN',
    'DELTOIDE': 'Deltoide',
    'EOGSIN': 'EOG_L',
    'EOGDX': 'EOG_R',
    'EOG-L': 'EOG_L',
    'EOG-R': 'EOG_R',
    'LOC': 'EOG_L',
    'ROC': 'EOG_R',
    # 'E1-M2'             : 'EOG_L',
    # 'E2-M2'             : 'EOG_R',
    'LEFTELECTROOCUL': 'EOG_L',
    'RIGHTELECTROOCU': 'EOG_R',
    'HEARTRATEVARIA': 'HR',
    'PLETH': 'Pleth',
    'CHEST': 'Thorax',
    'THORAX': 'Thorax',
    'TORACE': 'Thorax',
    'TORACICO': 'Thorax',
    'ABDOMEN': 'Abdomen',
    'ADDDOME': 'Abdomen',
    'ADDOME': 'Abdomen',
    'ABDO': 'Abdomen',
    'FLOW': 'Flow',
    'FLUSSO': 'Flow',
    'CANNULA': 'Cannula',
    'CANULA': 'Cannula',
    'FLATTENING': 'Cannula',
    'TERMISTORE': 'THE',
    'THERMISTOR': 'THE',
    'POSITION': 'Position',
    'POSIZIONE': 'Position',
    'OXSTATUS': 'STAT_SpO2',
    'STAT': 'STAT_SpO2',
    'SAO2': 'SpO2',
    'SPO2': 'SpO2',
    'SOUND': 'MIC',
    'LEG': 'Leg',
    'SX1': 'TIB_L1',
    'SX2': 'TIB_L2',
    'TIBSX': 'TIB_L',
    'TIBSIN': 'TIB_L',
    'DX1': 'TIB_R1',
    'DX2': 'TIB_R2',
    'TIBDX': 'TIB_R',
}
synonyms.update(**{
    left.upper() + right.upper(): f"{left}-{right}"
    for left in channel_labels_by_type['EEG']
    for right in channel_labels_by_type['EEG']
})

channel_type_by_label: Dict[str, str] = {}
for channel_type, channels in channel_labels_by_type.items():
    channel_type_by_label.update({ch: channel_type for ch in channels})

_conversion_map = {
    'uVuV': 1.0,
    'mVmV': 1.0,
    'uVmV': 10**-3,
    'mVuV': 10**3
}


def convert_units(in_unit, out_unit):
    return _conversion_map[in_unit+out_unit]



from typing import Dict, List

import numpy as np

from .channel_base import ChannelBase
from .label import Label


class Derivation(ChannelBase):
    operations = {
        '+': lambda l, r: l + r,
        '-': lambda l, r: l - r,
    }

    def __init__(self, left, right):
        self.left = left
        self.right = right
        self.label, operation = left.label.derive(right.label)
        self.op = self.operations[operation]

    def __getitem__(self, sli: slice) -> np.ndarray:
        """return a slice of the derivation"""
        left = self.left[sli]
        right = self.right[sli]
        return self.op(left, right)

    def from_dict(self, signals: Dict[Label, np.ndarray]) -> np.ndarray:
        left = self.left.from_dict(signals)
        right = self.right.from_dict(signals)
        return self.op(left, right)

    @property
    def channel_type(self):
        return self.left.channel_type

    @property
    def physical_dimension(self) -> str:
        return self.left.physical_dimension

    @property
    def num_samples_per_record(self):
        return self.left.num_samples_per_record

    @property
    def children(self) -> List[Label]:
        children = self.left.children
        children.extend(self.right.children)
        return children

    def derive(self, other: ChannelBase) -> ChannelBase:
        if not self.is_compatible(other):
            raise ValueError(f"cannot derive {self} and {other}")

        return Derivation(self, other)



from io import SEEK_SET
from struct import Struct
from typing import BinaryIO
from datetime import datetime

from .cached_property import cached_property
from .field import Field, normalize, serialize


class Header:
    fields = [
        Field('version', str, 8),
        Field('patient_id', str, 80),
        Field('recording_id', str, 80),
        Field('startdate', str, 8),
        Field('starttime', str, 8),
        Field('num_header_bytes', int, 8),
        Field('reserved', str, 44),
        Field('num_records', int, 8),
        Field('record_duration', float, 8),
        Field('num_channels', int, 4)
    ]
    # format_str = ''.join(str(size) + 's'
    #                      for _, _, size in fields_rec.bytesize)
    format_str = '8s80s80s8s8s8s44s8s8s4s'
    # _num_header_bytes = sum(field.size for field in fields)
    default_num_header_bytes = 256

    @property
    def version(self) -> str:
        return self._version

    @version.setter
    def version(self, v: str):
        self._version = normalize(str, v)

    @property
    def patient_id(self) -> str:
        return self._patient_id

    @patient_id.setter
    def patient_id(self, v: str):
        self._patient_id = normalize(str, v)

    @property
    def recording_id(self) -> str:
        return self._recording_id

    @recording_id.setter
    def recording_id(self, v: str):
        self._recording_id = normalize(str, v)

    @property
    def startdate(self) -> str:
        """Local start date of the recording"""
        return self._startdate

    @startdate.setter
    def startdate(self, v: str):
        self._startdate = normalize(str, v)

    @property
    def starttime(self) -> str:
        """Local start time of the recording"""
        return self._starttime

    @starttime.setter
    def starttime(self, v: str):
        self._starttime = normalize(str, v)

    @property
    def num_header_bytes(self) -> int:
        return self._num_header_bytes

    @num_header_bytes.setter
    def num_header_bytes(self, v: int):
        self._num_header_bytes = v

    @property
    def reserved(self) -> str:
        return self._reserved

    @reserved.setter
    def reserved(self, v: str):
        self._reserved = normalize(str, v)

    @property
    def num_records(self) -> int:
        """returns number of records"""
        return self._num_records

    @num_records.setter
    def num_records(self, v: int):
        self._num_records = v

    @property
    def record_duration(self) -> int:
        """returns seconds per record"""
        return self._record_duration

    @record_duration.setter
    def record_duration(self, v: int):
        self._record_duration = v

    @property
    def num_channels(self) -> int:
        """returns seconds per record"""
        return self._num_channels

    @num_channels.setter
    def num_channels(self, v: int):
        self._num_channels = v

    def __init__(self, *args, **kwargs):
        for k, v in kwargs.items():
            getattr(type(self), k).fset(self, v)

    @cached_property
    def startdatetime(self) -> datetime:
        datetime_str = f"{self.startdate}-{self.starttime}"
        try:
            return datetime.strptime(datetime_str, '%d.%m.%y-%H.%M.%S')
        except BaseException:
            # sometimes the day and month are switched
            return datetime.strptime(datetime_str, '%m.%d.%y-%H.%M.%S')

    @classmethod
    def read(cls, file: BinaryIO):
        data = file.read(cls.default_num_header_bytes)
        values = Struct(cls.format_str).unpack(data)
        named_content = {
            field.name: normalize(field.type, value)
            for field, value in zip(cls.fields, values)
        }
        return cls(**named_content)

    def write(self, file: BinaryIO):
        blob = [serialize(getattr(self, field.name), field.size)
                for field in self.fields]
        packed = Struct(self.format_str).pack(*blob)
        file.seek(0, SEEK_SET)
        file.write(packed)



from typing import List, Dict, Iterable
from datetime import datetime
from itertools import product
import numpy as np
from .blob import read_blob
from .label import Label
from .header import Header
from .channel import Channel


class Reader:
    def __init__(self, header: Header, channels: List[Channel]):
        """initialize reader with a filepath"""
        self.header = header
        self.basic_labels = [c.label for c in channels]
        self.channel_by_label = {c.label: c for c in channels}
        self.derivation_by_label = {
            k: v for k, v in self.channel_by_label.items()
        }
        self.compute_derivations()

    @classmethod
    def open(cls, filepath: str) -> 'Reader':
        with open(filepath, 'rb') as fp:
            header = Header.read(fp)
            channels = Channel.read(fp, header.num_channels)

        offset = header.num_header_bytes
        record_lengths = [c.num_samples_per_record for c in channels]
        blob_slices = read_blob(filepath, offset, record_lengths)
        for channel, blob_slice in zip(channels, blob_slices):
            channel.signal = blob_slice

        return cls(header, channels)

    def compute_derivations(self):
        channels = list(self.derivation_by_label.values())
        for left, right in product(channels, channels):
            if left is right or not left.is_compatible(right):
                continue

            derivation = left.derive(right)
            if derivation.label not in self.derivation_by_label:
                self.derivation_by_label[derivation.label] = derivation

    @property
    def labels(self):
        return list(self.derivation_by_label.keys())

    @property
    def duration(self) -> int:
        """returns recording duration in seconds"""
        return self.header.record_duration * self.header.num_records

    @property
    def startdatetime(self) -> datetime:
        """returns the time point of recording start"""
        return self.header.startdatetime

    def get_physical_samples(self, t0: float = 0.0, dt: float = None,
                             labels: List[str] = None) -> Dict[Label, np.ndarray]:  # noqa: E501
        """returns dict of samples by label from `t0` to `t0+dt`."""
        dt = dt or self.duration
        t1 = t0 + dt
        labels1 = list(map(Label, labels)) if labels else self.basic_labels
        required_labels = self.required_from_requested(labels1)
        rd = self.header.record_duration
        signals = {}
        for label in required_labels:
            channel = self.channel_by_label[label]
            sr = channel.num_samples_per_record / rd
            a = int(np.round(t0 * sr))
            b = int(np.round(t1 * sr))
            signals[label] = channel[a:b]

        return {
            ll: self.derivation_by_label[ll].from_dict(signals)
            for ll in labels1
        }

    def required_from_requested(self, labels: List[Label]) -> Iterable[Label]:
        """returns the labels required to construct the requested signals"""
        for label in labels:
            yield from self.derivation_by_label[label].children



from typing import Union
from collections import namedtuple

Field = namedtuple('Field', 'name, type, size')


def normalize(dtype, value: Union[str, bytes]):
    if isinstance(value, str):
        value = value.strip('\x00').strip(' ')
    elif isinstance(value, bytes):
        value = value.strip(b'\x00')
    else:
        AttributeError(f"Unknown type {type(value)}")

    if dtype is str:
        if isinstance(value, bytes):
            try:
                return str(value, 'ascii')
            except BaseException:
                return str(value, 'latin1')

    return dtype(value)


def serialize(value, size):
    format_str = '{:<%i}' % size if size else '{}'
    return bytes(format_str.format(value), 'latin1')



from os.path import join
import setuptools

__version__ = ''
exec(open(join('edfpy', 'version.py')).read())
requirements = open('requirements.txt').read().split('\n')

setuptools.setup(
    name='edfpy',
    version=__version__,
    packages=setuptools.find_packages(),
    author='Justus Schwabedal, John Snyder',
    maintainer='Justus Schwabedal',
    maintainer_email='jschwabedal@gmail.com',
    description="Lean EDF Reader",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)




import numpy as np
import matplotlib.pyplot as plt


def plot_physical_samples(fo, t0, dt):
    ax = None
    X = fo.get_physical_samples(t0, dt)
    plt.figure(figsize=(20, 10))
    for i, (c, x) in enumerate(X.items()):
        ax = plt.subplot(fo.num_channels, 1, 1+i, frameon=False, sharex=ax)
        t = np.linspace(t0, t0+dt, num=x.size)
        plt.plot(t, x, 'k-')
        plt.xticks(fontsize=20)
        plt.setp(ax.get_xticklabels(), visible=False)
        plt.yticks(fontsize=20)
        plt.ylabel(c, fontsize=20)
        plt.grid()
    plt.setp(ax.get_xticklabels(), visible=True)
    plt.xlabel('time (sec.)', fontsize=20)
    plt.xlim(t[0], t[-1])
    plt.tight_layout()
    plt.show()



from typing import List, Tuple

import numpy as np


class BlobSlice:
    def __init__(self, blob: np.memmap, locs: Tuple[int, int]):
        self.blob = blob
        self.block_size = locs[1] - locs[0]
        self.length = blob.shape[0] * self.block_size
        self.locs = slice(*locs)

    def __getitem__(self, sl: slice) -> np.ndarray:
        if sl.step and sl.step != 1:
            raise ValueError('slicing only with step width 1')

        q = self.block_size
        i = sl.start if sl.start else 0
        j = sl.stop if sl.stop else self.length
        j = self.length + j if j < 0 else j
        A = i // q
        B = int(np.ceil(j / q))
        a = i - A * q
        b = j - B * q or None
        block = self.blob[A:B, self.locs].flatten()
        return block[a:b]

    def __eq__(self, other):
        return self[:] == other

    def __repr__(self):
        return f"BlobSlice({self[:]})"


def read_blob(file, offset: int, record_lengths: List[int]):
    memarr = np.memmap(file, dtype='<i2',  # type: ignore
                       mode='r', offset=offset)
    pos = np.cumsum([0] + record_lengths).astype(int)
    memarr.shape = (-1, pos[-1])
    locs = zip(pos[:-1], pos[1:])
    return [BlobSlice(memarr, loc) for loc in locs]


def write_blob(file, arrs: List[np.ndarray], record_lengths: List[int]):
    reshaped = [arr.reshape((-1, n)) for arr, n in zip(arrs, record_lengths)]
    blob = np.concatenate(reshaped, axis=1).tobytes()
    file.write(blob)



from typing import List, Dict

import numpy as np

from .label import Label
from .field import normalize


class ChannelBase:
    @property
    def label(self) -> Label:
        return self._label

    @label.setter
    def label(self, v: str):
        n = normalize(Label, v)
        self._label = n

    @property
    def children(self) -> List[Label]:
        raise NotImplementedError

    @property
    def physical_dimension(self) -> str:
        raise NotImplementedError

    @property
    def num_samples_per_record(self) -> int:
        raise NotImplementedError

    def is_compatible(self, other: 'ChannelBase') -> bool:
        compat_label = self.label.is_compatible(other.label)
        same_units = self.physical_dimension == other.physical_dimension
        same_sr = self.num_samples_per_record == other.num_samples_per_record
        return same_units and same_sr and compat_label

    def from_dict(self, signals_dict: Dict[Label, np.ndarray]) -> np.ndarray:
        raise NotImplementedError
